/** @file sdl_socket.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2013.12.06

	@brief [ sdl socket ]

	@note
	This code is based on win32_socket.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
	Date   : 2006.08.27 -
*/

#include "emu.h"
#include "vm/vm.h"

#define USE_SDL_EVENT

#ifndef USE_SDL_EVENT
static EMU *emu;
// for callback timer
typedef struct {
	int ch;
	SDL_TimerID id;
} st_timer;
// callback
static Uint32 on_connected(Uint32 interval, void *param);
static Uint32 on_disconnected(Uint32 interval, void *param);
static Uint32 on_writeable(Uint32 interval, void *param);
static Uint32 on_readable(Uint32 interval, void *param);
#endif

void EMU::EMU_SOCKET()
{
	// init sockets
	for(int i = 0; i < SOCKET_MAX; i++) {
//		soc[i] = INVALID_SOCKET;
		tcp_socket[i] = NULL;
		udp_socket[i] = NULL;
		is_tcp[i] = false;
		is_server[i] = false;
		socket_delay[i] = 0;
		recv_buffer[i] = NULL;
		recv_r_ptr[i] = recv_w_ptr[i] = 0;
	}
	socket_set = NULL;
	socket_nums = 0;
}

void EMU::initialize_socket()
{
#ifndef USE_SDL_EVENT
	emu = this;
#endif

	// init sdl net
	if(SDLNet_Init()==-1) {
		out_logf(LOG_ERROR, _T("SDLNet_Init: %s"), SDLNet_GetError());
	}
	socket_set = SDLNet_AllocSocketSet(SOCKET_MAX);
	if(!socket_set) {
		out_logf(LOG_ERROR, _T("SDLNet_AllocSocketSet: %s"), SDLNet_GetError());
	}
}

void EMU::release_socket()
{
	// release sockets
	for(int i = 0; i < SOCKET_MAX; i++) {
//		if(soc[i] != INVALID_SOCKET) {
//			shutdown(soc[i], 2);
//			closesocket(soc[i]);
//		}
		SDLNet_TCP_Close(tcp_socket[i]);
		SDLNet_UDP_Close(udp_socket[i]);
		tcp_socket[i] = NULL;
		udp_socket[i] = NULL;

		delete[] recv_buffer[i];
	}

	SDLNet_FreeSocketSet(socket_set);
	socket_set = NULL;

	// release sdl net
	SDLNet_Quit();
}

void EMU::socket_connected(int ch)
{
	// network is connected
	out_debugf(_T("connected. ch=%d"), ch);
	vm->network_connected(ch);
}

void EMU::socket_disconnected(int ch)
{
	// network is disconnected
//	if(!socket_delay[ch]) {
//		socket_delay[ch] = 1;//56;
//	}
	out_debugf(_T("disconnected. ch=%d"), ch);
}

void EMU::socket_writeable(int ch)
{
	// network is writeable
	out_debugf(_T("writeable. ch=%d"), ch);
	vm->network_writeable(ch);
}

void EMU::socket_readable(int ch)
{
	// network is readable
	out_debugf(_T("readable. ch=%d"), ch);
	vm->network_readable(ch);
}

void EMU::update_socket()
{
	// check arrived packets
	if (socket_nums > 0) {
		int retval = SDLNet_CheckSockets(socket_set, 0);
		if (retval < 0) {
			// error
			out_logf(LOG_ERROR, _T("SDLNet_CheckSockets: %s"), SDLNet_GetError());
		} else if (retval > 0) {
			// data arrived
			for(int i = 0; i < SOCKET_MAX; i++) {
				if (is_tcp[i]) {
					if (SDLNet_SocketReady(tcp_socket[i])) {
						if (is_server[i]) {
							socket_accept(i);
						} else {
							recv_data(i);
							socket_readable(i);
						}
					}
				} else {
					if (SDLNet_SocketReady(udp_socket[i])) {
						recv_data(i);
						socket_readable(i);
					}
				}
			}
		}
	}
	// buffer copy
	for(int i = 0; i < SOCKET_MAX; i++) {
		if(recv_r_ptr[i] < recv_w_ptr[i]) {
			// get buffer
			int size0, size1;
			uint8* buf0 = vm->get_recvbuffer0(i, &size0, &size1);
			uint8* buf1 = vm->get_recvbuffer1(i);

			int size = recv_w_ptr[i] - recv_r_ptr[i];
			if(size > size0 + size1) {
				size = size0 + size1;
			}
			char* src = &recv_buffer[i][recv_r_ptr[i]];
			recv_r_ptr[i] += size;

			if(size <= size0) {
				memcpy(buf0, src, size);
			}
			else {
				memcpy(buf0, src, size0);
				memcpy(buf1, src + size0, size - size0);
			}
			vm->inc_recvbuffer_ptr(i, size);
		}
		else if(socket_delay[i] != 0) {
			if(--socket_delay[i] == 0) {
				vm->network_disconnected(i);
			}
		}
	}
}

void EMU::socket_accept(int ch)
{
	int new_ch = -1;

	TCPsocket new_tcpsock;

	new_tcpsock=SDLNet_TCP_Accept(tcp_socket[ch]);
	if (new_tcpsock == NULL) {
		out_logf(LOG_ERROR, _T("SDLNet_TCP_Accept: ch=%d %s"), ch, SDLNet_GetError());
		return;
	}

	// find empty socket
	new_ch = get_socket_channel();
	if (new_ch < 0) {
		SDLNet_TCP_Close(new_tcpsock);
		return;
	}

	int numused = SDLNet_TCP_AddSocket(socket_set, new_tcpsock);
	if(numused == -1) {
		SDLNet_TCP_Close(new_tcpsock);
		out_logf(LOG_ERROR, _T("SDLNet_TCP_AddSocket: new_ch=%d %s"), new_ch, SDLNet_GetError());
		return;
	}

	IPaddress *remote_ip = SDLNet_TCP_GetPeerAddress(new_tcpsock);
	if (remote_ip) {
		out_debugf(_T("accepted from 0x%08x"), remote_ip->host);
	}

	tcp_socket[new_ch] = new_tcpsock;
	is_tcp[new_ch] = is_tcp[ch];
	socket_nums++;

	vm->network_accepted(ch, new_ch);

	out_debugf(_T("socket_accept ok. ch=%d is_tcp=%s new_ch=%d is_tcp=%s"), ch, is_tcp[ch] ? _T("true") : _T("false"), new_ch, is_tcp[new_ch] ? _T("true") : _T("false"));

	// TODO: post SDL event
	// note that i'll fire the writeable event only, won't fire the connected event.
#ifndef USE_SDL_EVENT
	st_timer *timer2 = new st_timer;
	timer2->ch = ch;
	timer2->id = SDL_AddTimer(32, on_writeable, (void *)timer2);
#else
	SDL_Event evt;
	evt.type = SDL_USEREVENT_SOCKET;
	evt.user.code = SDL_USEREVENT_SOCKET_WRITEABLE;
	evt.user.data1 = (void *)new_ch;
	evt.user.data2 = NULL;
	SDL_PushEvent(&evt);
#endif
	return;
}

bool EMU::init_socket_tcp(int ch, bool server)
{
	disconnect_socket(ch);

	// reserve channel for tcp connection
	is_tcp[ch] = true;

	recv_r_ptr[ch] = recv_w_ptr[ch] = 0;
	return true;
}

bool EMU::init_socket_udp(int ch, bool server)
{
	disconnect_socket(ch);

	// reserve channel for udp connection
	is_tcp[ch] = false;

	if (!server) {
		// client
		UDPsocket sock;

		sock = SDLNet_UDP_Open(0);
		if(!sock) {
			out_logf(LOG_ERROR, _T("SDLNet_UDP_Open: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		int numused = SDLNet_UDP_AddSocket(socket_set, sock);
		if(numused == -1) {
			SDLNet_UDP_Close(sock);
			out_logf(LOG_ERROR, _T("SDLNet_UDP_AddSocket: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		socket_nums++;
		udp_socket[ch] = sock;
		is_server[ch] = server;

		// TODO: post SDL event
#ifndef USE_SDL_EVENT
		st_timer *timer2 = new st_timer;
		timer2->ch = ch;
		timer2->id = SDL_AddTimer(32, on_writeable, (void *)timer2);
#else
		SDL_Event evt;
		evt.type = SDL_USEREVENT_SOCKET;
		evt.user.code = SDL_USEREVENT_SOCKET_WRITEABLE;
		evt.user.data1 = (void *)ch;
		evt.user.data2 = NULL;
		SDL_PushEvent(&evt);
#endif
	}

	recv_r_ptr[ch] = recv_w_ptr[ch] = 0;
	return true;
}

bool EMU::connect_socket(int ch, IPaddress *ip, bool server)
{
	if (is_tcp[ch]) {
		TCPsocket sock;

		if (server == false && (ip->host == INADDR_NONE || ip->host == INADDR_ANY)) {
			out_log(LOG_ERROR, _T("Cannot create socket as the server."));
			return false;
		}

		sock = SDLNet_TCP_Open(ip, server ? SDL_TRUE : SDL_FALSE);
		if(!sock) {
			out_logf(LOG_ERROR, _T("SDLNet_TCP_Open: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		int numused = SDLNet_TCP_AddSocket(socket_set, sock);
		if(numused == -1) {
			SDLNet_TCP_Close(sock);
			out_logf(LOG_ERROR, _T("SDLNet_TCP_AddSocket: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		tcp_socket[ch] = sock;
		socket_nums++;
	} else {
		if (!server) {
			// because client setting was set by init_socket_udp
			return true;
		}

		UDPsocket sock;

		sock = SDLNet_UDP_Open(ip->port);
		if(!sock) {
			out_logf(LOG_ERROR, _T("SDLNet_UDP_Open: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		int channel = SDLNet_UDP_Bind(sock, -1, ip);
		if (channel == -1) {
			SDLNet_UDP_Close(sock);
			out_logf(LOG_ERROR, _T("SDLNet_UDP_Bind: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		int numused = SDLNet_UDP_AddSocket(socket_set, sock);
		if(numused == -1) {
			SDLNet_UDP_Close(sock);
			out_logf(LOG_ERROR, _T("SDLNet_UDP_AddSocket: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		udp_socket[ch] = sock;
		socket_nums++;
	}

	is_server[ch] = server;

	// TODO: post SDL event
#ifndef USE_SDL_EVENT
	st_timer *timer1 = new st_timer;
	timer1->ch = ch;
	timer1->id = SDL_AddTimer(16, on_connected, (void *)timer1);
	st_timer *timer2 = new st_timer;
	timer2->ch = ch;
	timer2->id = SDL_AddTimer(32, on_writeable, (void *)timer2);
#else
	SDL_Event evt;
	evt.type = SDL_USEREVENT_SOCKET;
	evt.user.code = SDL_USEREVENT_SOCKET_CONNECTED;
	evt.user.data1 = (void *)ch;
	evt.user.data2 = NULL;
	SDL_PushEvent(&evt);

	evt.type = SDL_USEREVENT_SOCKET;
	evt.user.code = SDL_USEREVENT_SOCKET_WRITEABLE;
	evt.user.data1 = (void *)ch;
	evt.user.data2 = NULL;
	SDL_PushEvent(&evt);
#endif
	return true;
}

bool EMU::connect_socket(int ch, uint32 ipaddr, int port, bool server)
{
	IPaddress ip;

	ip.host = (Uint32)ipaddr;
	ip.port = (Uint16)port;

	return connect_socket(ch, &ip, server);
}

bool EMU::connect_socket(int ch, const _TCHAR *hostname, int port, bool server)
{
	IPaddress ip;

	if(!get_ipaddr(hostname, port, &ip)) {
		return false;
	}

	return connect_socket(ch, &ip, server);
}

bool EMU::get_ipaddr(const _TCHAR *hostname, int port, IPaddress *ip)
{
#ifdef _UNICODE
	char host[_MAX_PATH];
	wcstombs(host, hostname, _MAX_PATH);
#else
	const char *host = hostname;
#endif

	if(SDLNet_ResolveHost(ip, (const char *)host, (Uint16)port) == -1) {
		// unresolved hostname to ipaddr
		out_logf(LOG_ERROR, _T("SDLNet_ResolveHost: %s"), SDLNet_GetError());
		return false;
	}

	return true;
}

void EMU::disconnect_socket(int ch)
{
	if (is_tcp[ch]) {
		if (tcp_socket[ch]) {
			SDLNet_TCP_Close(tcp_socket[ch]);
			SDLNet_TCP_DelSocket(socket_set, tcp_socket[ch]);
			tcp_socket[ch] = NULL;
			socket_nums--;
			vm->network_disconnected(ch);
		}
	} else {
		if (udp_socket[ch]) {
			SDLNet_UDP_Close(udp_socket[ch]);
			SDLNet_UDP_DelSocket(socket_set, udp_socket[ch]);
			udp_socket[ch] = NULL;
			socket_nums--;
			vm->network_disconnected(ch);
		}
	}

	// TODO: post SDL event
#ifndef USE_SDL_EVENT
	st_timer *timer = new st_timer;
	timer->ch = ch;
	timer->id = SDL_AddTimer(16, on_disconnected, (void *)timer);
#else
	SDL_Event evt;
	evt.type = SDL_USEREVENT_SOCKET;
	evt.user.code = SDL_USEREVENT_SOCKET_DISCONNECTED;
	evt.user.data1 = (void *)ch;
	evt.user.data2 = NULL;
	SDL_PushEvent(&evt);
#endif
}


int EMU::get_socket_channel()
{
	int new_ch = -1;
	for(int i = 0; i < SOCKET_MAX; i++) {
		if (tcp_socket[i] == NULL && udp_socket[i] == NULL) {
			new_ch = i;
			break;
		}
	}
	if (new_ch < 0) {
		out_log(LOG_ERROR, _T("Socket is full."));
	}
	return new_ch;
}

bool EMU::listen_socket(int ch)
{
	return false;
}

void EMU::send_data_tcp(int ch)
{
	if(is_tcp[ch]) {
		send_data(ch);
	}
}

void EMU::send_data_udp(int ch, uint32 ipaddr, int port)
{
	if(!is_tcp[ch]) {
		udp_packet.address.host = (Uint32)ipaddr;
		udp_packet.address.port = (Uint16)port;
		udp_packet.channel = -1;

		send_data(ch);
	}
}

void EMU::send_data_udp(int ch, const _TCHAR *hostname, int port)
{
	IPaddress ip;

	if (!get_ipaddr(hostname, port, &ip)) {
		return;
	}

	send_data_udp(ch, (uint32)ip.host, ip.port);
}

void EMU::send_data(int ch)
{
	// loop while send buffer is not empty
	while(1) {
		// get send buffer and data size
		int size;
		uint8* buf = vm->get_sendbuffer(ch, &size);

		if(!size) {
			return;
		}
		if(is_tcp[ch]) {
			int sent_size = SDLNet_TCP_Send(tcp_socket[ch], (const void *)buf, size);
			if(sent_size < size) {
				// error
				disconnect_socket(ch);
				socket_disconnected(ch);
				return;
			}
			size = sent_size;
		} else {
			udp_packet.data = (Uint8 *)buf;
			udp_packet.maxlen = size;
			udp_packet.len = size;
			int sent_packet = SDLNet_UDP_Send(udp_socket[ch], udp_packet.channel, &udp_packet);
			if(sent_packet <= 0) {
				// error
				disconnect_socket(ch);
				socket_disconnected(ch);
				return;
			}
		}
		vm->inc_sendbuffer_ptr(ch, size);
	}
}

void EMU::recv_data(int ch)
{
	if (recv_buffer[ch] == NULL) {
		recv_buffer[ch] = new char[SOCKET_BUFFER_MAX];
	}

	if(is_tcp[ch]) {
		int size = SOCKET_BUFFER_MAX - recv_w_ptr[ch];
		char* buf = &recv_buffer[ch][recv_w_ptr[ch]];
		int recv_size = SDLNet_TCP_Recv(tcp_socket[ch], (void *)buf, size);
		if(recv_size <= 0) {
			disconnect_socket(ch);
			socket_disconnected(ch);
			return;
		}
		recv_w_ptr[ch] += recv_size;
	}
	else {
		UDPpacket packet;
		int size = SOCKET_BUFFER_MAX - recv_w_ptr[ch];
		char *buf = &recv_buffer[ch][recv_w_ptr[ch]];

		packet.address.host = 0;
		packet.address.port = 0;
		packet.channel = 0;
		packet.maxlen = size - 8;
		packet.len = 0;
		packet.data = (Uint8 *)(buf + 8);

		if(size < 8) {
			return;
		}
		int recv_packet = SDLNet_UDP_Recv(udp_socket[ch], &packet);
//		if((size = recvfrom(soc[ch], buf + 8, size - 8, 0, (struct sockaddr *)&addr, &len)) == SOCKET_ERROR) {
		if(recv_packet < 0) {
			disconnect_socket(ch);
			socket_disconnected(ch);
			return;
		}
//		size += 8;
		buf[0] = packet.len >> 8;
		buf[1] = packet.len;
		buf[2] = (char)packet.address.host;
		buf[3] = (char)(packet.address.host >> 8);
		buf[4] = (char)(packet.address.host >> 16);
		buf[5] = (char)(packet.address.host >> 24);
		buf[6] = (char)packet.address.port;
		buf[7] = (char)(packet.address.port >> 8);
		recv_w_ptr[ch] += packet.len;
	}
}

void *EMU::get_socket(int ch)
{
	if (is_tcp[ch]) {
		return tcp_socket[ch];
	} else {
		return udp_socket[ch];
	}
}

#ifndef USE_SDL_EVENT
//
// callback on timer
//
static Uint32 on_connected(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->socket_connected(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
static Uint32 on_disconnected(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->socket_disconnected(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
static Uint32 on_writeable(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->socket_writeable(timer->ch);
	emu->send_data(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
static Uint32 on_readable(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->recv_data(timer->ch);
	emu->socket_readable(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
#endif // USE_SDL_EVENT