/** @file hd46505.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2007.02.08 -

	@note
	Modified by Sasaji at 2011.12.07

	@brief [ HD46505 ]
*/

#ifndef _HD46505_H_
#define _HD46505_H_

#include "vm.h"
#include "../emu.h"
#include "device.h"

#define SIG_HD46505_CPU_POWER	1
#define SIG_HD46505_CHR_CLOCKS	2

/**
	@brief HD46505 - CRT Controller
*/
class HD46505 : public DEVICE
{
private:
	// output signals
	outputs_t outputs_disp;
	outputs_t outputs_vblank;
	outputs_t outputs_vsync;
	outputs_t outputs_hsync;

	uint8 regs[18];
	int ch;
	bool timing_changed;

	int cpu_clocks;
	int chr_clocks;
#ifdef HD46505_HORIZ_FREQ
	int horiz_freq, next_horiz_freq;
#endif
	double frames_per_sec;

	int hz_total, hz_disp;
	int hs_start, hs_end;

	int vt_total, vt_disp;
	int vt_total_per_frame;
	int vs_start, vs_end;
	int v_count;

	int disp_end_clock;
	int hs_start_clock, hs_end_clock;

	bool display, vblank, vsync, hsync;

	int register_id[3];

#pragma pack(1)
	struct vm_state_st {
		uint8 regs[18];
		int ch;
		int cpu_clocks;
		int chr_clocks;
		char reserved[2];
		// version 2
		int register_id[3];
		int v_count;
	} vm_state;
#pragma pack()

	void set_display(bool val);
	void set_vblank(bool val);
	void set_vsync(bool val);
	void set_hsync(bool val);

public:
	HD46505(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("HD46505");
		init_output_signals(&outputs_disp);
		init_output_signals(&outputs_vblank);
		init_output_signals(&outputs_vsync);
		init_output_signals(&outputs_hsync);
	}
	~HD46505() {}

	// common functions
	void initialize();
	void reset();
	void write_io8(uint32 addr, uint32 data);
	uint32 read_io8(uint32 addr);
	void event_pre_frame();
	void event_frame();
	void event_vline(int v, int clock);
	void event_callback(int event_id, int err);
	void update_timing(int new_clocks, double new_frames_per_sec, int new_lines_per_frame);
	void write_signal(int id, uint32 data, uint32 mask);

	// unique function
	void set_context_disp(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_disp, device, id, mask);
	}
	void set_context_vblank(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_vblank, device, id, mask);
	}
	void set_context_vsync(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_vsync, device, id, mask);
	}
	void set_context_hsync(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_hsync, device, id, mask);
	}
#ifdef HD46505_HORIZ_FREQ
	void set_horiz_freq(int freq) {
		next_horiz_freq = freq;
	}
#endif
	uint8* get_regs() {
		return regs;
	}
	int* get_vt_total_ptr() {
		return &vt_total;
	}
	int* get_vt_count_ptr() {
		return &v_count;
	}
	int* get_vt_disp_ptr() {
		return &vt_disp;
	}

	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);
};

#endif

