/** @file floppy.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.11.05 -

	@brief [ floppy drive ]
*/

#include "floppy.h"
#include "mc6843.h"
#include "mb8866.h"
#include "../disk.h"

#if defined(_DEBUG_FLOPPY) || defined(_DEBUG_MB6843) || defined(_DEBUG_MB8866)
#define OUT_DEBUG emu->out_debugf
#else
#define OUT_DEBUG dummyf
#endif

#define EVENT_IRQ			0
#define EVENT_DRQ			1

#define EVENT_MOTOR_TIMEOUT	2
#define EVENT_INDEXHOLE_ON	3
#define EVENT_INDEXHOLE_OFF	4

#define EVENT_READY_ON_0	5
#define EVENT_READY_ON_1	6
#define EVENT_READY_ON_2	7
#define EVENT_READY_ON_3	8
#define EVENT_READY_ON_4	9
#define EVENT_READY_ON_5	10
#define EVENT_READY_ON_6	11
#define EVENT_READY_ON_7	12
#define EVENT_MOTOR_OFF_0	13
#define EVENT_MOTOR_OFF_1	14
#define EVENT_MOTOR_OFF_2	15
#define EVENT_MOTOR_OFF_3	16
#define EVENT_MOTOR_OFF_4	17
#define EVENT_MOTOR_OFF_5	18
#define EVENT_MOTOR_OFF_6	19
#define EVENT_MOTOR_OFF_7	20

// index hole (usually 300rpm -> 200ms)
#define DELAY_INDEX_HOLE	  200000
//#define DELAY_INDEX_HOLE	   50000
// index hole (2HD Type) (usually 360rpm -> 167ms)
#define DELAY_INDEX_HOLE_H	  166667

// ready on delay time after motor on (us)
#define DELAY_READY_ON		 1000000
// ready on delay time after motor on (us) (2HD Type) 
#define DELAY_READY_ON_H	  500000

// motor warmup time (us)
#define DELAY_MOTOR_WARMUP	  500000
// motor off delay time (us)
#define DELAY_MOTOR_OFF_3	 3000000
#define DELAY_MOTOR_OFF_5S	 5000000
#define DELAY_MOTOR_OFF_5	11000000
#define DELAY_MOTOR_TIMEOUT	60000000

#define DRIVE_MASK	(MAX_DRIVE - 1)


void FLOPPY::cancel_my_event(int event_no) {
	if(register_id[event_no] != -1) {
		cancel_event(register_id[event_no]);
		register_id[event_no] = -1;
	}
}
void FLOPPY::register_my_event(int event_no, int wait) {
	if(register_id[event_no] != -1) {
		cancel_event(register_id[event_no]);
		register_id[event_no] = -1;
	}
	register_event(this, event_no, wait, false, &register_id[event_no]);
}

void FLOPPY::initialize()
{
	// setup/reset floppy drive
	for(int i = 0; i < MAX_DRIVE; i++) {
		disk[i] = new DISK();
	}
	for(int i = 0; i < MAX_DRIVE; i++) {
		fdc[i].side = 0;
		fdc[i].track = 0;
		fdc[i].index = 0;
		fdc[i].access = false;
		fdc[i].ready = 0;
		fdc[i].motor_warmup = 0;
		fdc[i].head_loading = 0;
	}
	ignore_crc = false;

	sidereg = 0;
	sectorcnt = 0;

	head_load = 0;

	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		register_id[i] = -1;
	}

	wav_enable = 0;
	wav_loaded_at_first = false;

	for(int ty = 0; ty < 2; ty++) {
		for(int ft = 0; ft < 2; ft++) {
			wav_data[ft][ty] = NULL;
			wav_size[ft][ty] = 0;
		}
		wav_play_pos[ty] = 0;
		wav_play[ty] = 0;
	}
	_tcscpy(wav_file[0][0], _T("fddseek3.wav"));
	_tcscpy(wav_file[0][1], _T("fddmotor3.wav"));
	_tcscpy(wav_file[1][0], _T("fddseek5.wav"));
	_tcscpy(wav_file[1][1], _T("fddmotor5.wav"));

	register_frame_event(this);
}

void FLOPPY::reset()
{
	load_wav();

	for(int i = 0; i < MAX_DRIVE; i++) {
		fdc[i].track = 0;
		fdc[i].index = 0;
		fdc[i].access = false;
		fdc[i].ready = 0;
		fdc[i].motor_warmup = 0;
		fdc[i].head_loading = 0;
	}

	// events were already canceled by EVENT::reset()
	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		register_id[i] = -1;
	}

	drv_num[0] = 0;
	drv_num[1] = 0;
	drvsel[0] = 0;
	drvsel[1] = 0;

	index_hole = 0;

	set_drive_speed();

	density = 0;

	fdd5outreg[0] = 0x81;
	fdd5outreg[1] = 0x81;
	irqflg = false;
	irqflgprev = irqflg;
	drqflg = false;
	drqflgprev = drqflg;

	ignore_write = false;

	wav_play[0] = 0;
	wav_play[1] = 0;

	if (config.io_port & 0x03) {
		register_my_event(EVENT_INDEXHOLE_ON, delay_index_hole);	// index hole
		register_my_event(EVENT_MOTOR_TIMEOUT, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)
	}
}

void FLOPPY::cancel_my_events()
{
	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		cancel_my_event(i);
	}
}

void FLOPPY::release()
{
	// release d88 handler
	for(int i = 0; i < MAX_DRIVE; i++) {
		delete disk[i];
	}
	for(int i = 0; i < 2; i++) {
		for(int j = 0; j < 2; j++) {
			delete[] wav_data[i][j];
		}
	}
}

void FLOPPY::load_wav()
{
	// load wav file
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];
	wav_header_t	 head;
	wav_fmt_chank_t  fmt;
	wav_data_chank_t data;
	size_t			 data_len;
	size_t           data_max[2];
	PARSEWAV_UTIL    wavu;

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	data_max[0] = (sample_rate >> 3);	// 0.125sec (seek)
	data_max[1] = (sample_rate / 5);	// 0.2sec (moror)
	for (int j=0; j<4; j++) {
		if ((wav_enable & (1 << j)) == 0) {
			wav_data[j >> 1][j & 1] = new uint8[data_max[j & 1]];
			memset(wav_data[j >> 1][j & 1], 128, sizeof(uint8) * data_max[j & 1]);
		}
	}

	FILEIO* fre = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		for (int j=0; j<4; j++) {
			if ((wav_enable & (1 << j)) == 0) {
				UTILITY::stprintf(file_path, _MAX_PATH, _T("%s%s"), app_path, wav_file[j >> 1][j & 1]);
				if(fre->Fopen(file_path, FILEIO_READ_BINARY)) {
					if (wavu.CheckWavFormat(fre, &head, &fmt, &data, &data_len) == 0) {
						wav_size[j >> 1][j & 1] = (int)wavu.ReadWavData(fre, &fmt, data_len, wav_data[j >> 1][j & 1], sample_rate, 8, data_max[j & 1]);
						wav_enable |= (1 << j);
						emu->out_logf(LOG_INFO, (CT)_("%s was loaded."),wav_file[j >> 1][j & 1]);
					}
					fre->Fclose();
				}
			}
		}
	}
	delete fre;

	for (int j=0; j<4; j++) {
		if ((wav_enable & (1 << j)) == 0) {
			wav_size[j >> 1][j & 1] = 0;
			delete[] wav_data[j >> 1][j & 1];
			wav_data[j >> 1][j & 1] = NULL;
			if (!wav_loaded_at_first) {
				OUT_DEBUG((CT)_("%s couldn't be loaded."),wav_file[j >> 1][j & 1]);
			}
		}
	}
	wav_loaded_at_first = true;
}

void FLOPPY::write_io8(uint32 addr, uint32 data)
{
	if (IOPORT_USE_5FDD) {
		// 5inch
		if (config.fdd_type == 3 && NOW_L3_MODE) {
			addr = (addr & ~0x10) | (0x10 - (addr & 0x10));
		}
		switch(addr & 0x1f) {
			case 0x0:
			case 0x1:
			case 0x2:
				// FDC access ok
				if (config.fdd_type != 3 || (config.fdd_type == 3 && (drvsel[0] & 0x80) == 0)) {
					OUT_DEBUG(_T("fddw a:%05x d:%02x (%02x)"),addr,data,(~data) & 0xff);
					d_fdc5->write_io8(addr & 0xf, data);
				} else {
					OUT_DEBUG(_T("fddw a:%05x d:%02x (%02x) BLOCKED"),addr,data,(~data) & 0xff);
				}
				break;
			case 0x3:
				// FDC access ok or DRQ on
				if (config.fdd_type != 3 || (config.fdd_type == 3 && ((drvsel[0] & 0x80) == 0 || (fdd5outreg[0] & 0x80) == 0))) {
//					OUT_DEBUG(_T("fddw a:%05x d:%02x (%02x)"),addr,data,(~data) & 0xff);
					d_fdc5->write_io8(addr & 0xf, data);
				} else {
					OUT_DEBUG(_T("fddw a:%05x d:%02x (%02x) BLOCKED"),addr,data,(~data) & 0xff);
				}
				break;
			case 0x4:
				set_drive_select(1, data);
				OUT_DEBUG(_T("fddw a:%05x d:%02x"),addr,data);
				break;
			case 0x8:
				// Send HALT signal if DRQ signal on FDC is false. 
				if (!drqflg) {
					OUT_DEBUG(_T("fddw a:%05x d:%02x drq:%02x"),addr,data,drqflg);
					d_board->write_signal(SIG_CPU_HALT, SIG_HALT_FD_MASK, SIG_HALT_FD_MASK);
				}
				break;
			case 0xc:
				break;

			/* FDC (ex 2D) */
			case 0x10:
			case 0x11:
			case 0x12:
			case 0x13:
				// FDC access
				if (config.fdd_type == 3) {
					OUT_DEBUG(_T("fddw a:%05x d:%02x (%02x)"),addr,data,(~data) & 0xff);
					d_fdc5ex->write_io8(addr & 0xf, data);
				}
				break;
			case 0x14:
				set_drive_select(2, data);
				OUT_DEBUG(_T("fddw a:%05x d:%02x"),addr,data);
				break;
		}
	} else if (IOPORT_USE_3FDD) {
		// 3inch
		set_drive_select(0, data);
	}
}

uint32 FLOPPY::read_io8(uint32 addr)
{
	uint32 data = 0xff;

	if (IOPORT_USE_5FDD) {
		// 5inch
		if (config.fdd_type == 3 && NOW_L3_MODE) {
			addr = (addr & ~0x10) | (0x10 - (addr & 0x10));
		}
		switch(addr & 0x1f) {
			case 0x0:
			case 0x1:
				// FDC access ok
				if (config.fdd_type != 3 || (config.fdd_type == 3 && (drvsel[0] & 0x80) == 0)) {
					data = d_fdc5->read_io8(addr & 0xf);
//					OUT_DEBUG(_T("fddr a:%05x d:%02x (%02x)"),addr,data,(~data) & 0xff);
				} else {
					OUT_DEBUG(_T("fddr a:%05x d:%02x (%02x) BLOCKED"),addr,data,(~data) & 0xff);
				}
				break;
			case 0x2:
				// FDC access ok
				if (config.fdd_type != 3 || (config.fdd_type == 3 && (drvsel[0] & 0x80) == 0)) {
					data = d_fdc5->read_io8(addr & 0xf);
					OUT_DEBUG(_T("fddr a:%05x d:%02x (%02x)"),addr,data,(~data) & 0xff);
				} else {
					// When FDC access mask is ON, send DRQ signal
					data = (fdd5outreg[0] & 0x80);
//					OUT_DEBUG(_T("fddr a:%05x d:%02x (%02x) Wait DRQ"),addr,data);
				}
				break;
			case 0x3:
				// FDC access ok or DRQ on
				if (config.fdd_type != 3 || (config.fdd_type == 3 && ((drvsel[0] & 0x80) == 0 || (fdd5outreg[0] & 0x80) == 0))) {
					data = d_fdc5->read_io8(addr & 0xf);
//					OUT_DEBUG(_T("fddr a:%05x d:%02x (%02x)"),addr,data,(~data) & 0xff);
				} else {
//					OUT_DEBUG(_T("fddr a:%05x d:%02x (%02x) BLOCKED"),addr,data,(~data) & 0xff);
				}
				break;
			case 0x4:
				data = fdd5outreg[0];
				if (config.fdd_type == 3) data |= 0x80;
				OUT_DEBUG(_T("fddr a:%05x d:%02x"),addr,data);
				break;
			case 0xc:
				// Type
				if (config.fdd_type == 3) {
					// always zero
					data = 0x00;
				}
				break;

			/* FDC (ex 2D) */
			case 0x10:
			case 0x11:
			case 0x12:
			case 0x13:
				// FDC access
				if (config.fdd_type == 3) {
					data = d_fdc5ex->read_io8(addr & 0xf);
				}
				break;
			case 0x14:
				if (config.fdd_type == 3) {
					data = fdd5outreg[1];
					OUT_DEBUG(_T("fddr a:%05x d:%02x"),addr,data);
				}
				break;
		}
	} else if (IOPORT_USE_3FDD) {
		// 3inch
		data = drvsel[0];
	}

	return data;
}

void FLOPPY::write_signal(int id, uint32 data, uint32 mask)
{
	int fdcnum = (id >> 16);
	id &= 0xffff;

	int ndrv_num = drv_num[fdcnum];

	switch (id) {
		case SIG_FLOPPY_ACCESS:
			fdc[ndrv_num].access = (data & 1) ? true : false;
			break;
		case SIG_FLOPPY_WRITE:
			if(!ignore_write && fdc[ndrv_num].index < disk[ndrv_num]->sector_size) {
				disk[ndrv_num]->sector[fdc[ndrv_num].index] = (data & mask);
				fdc[ndrv_num].index++;
			}
			break;
		case SIG_FLOPPY_WRITE_TRACK:
			if(!ignore_write && fdc[ndrv_num].index < disk[ndrv_num]->track_size) {
				disk[ndrv_num]->track[fdc[ndrv_num].index] = (data & mask);
				fdc[ndrv_num].index++;
			}
			break;
		case SIG_FLOPPY_WRITEDELETE:
			if (!ignore_write) {
				disk[ndrv_num]->deleted = (data & 1) ? 0x10 : 0;
			}
			break;
		case SIG_FLOPPY_STEP:
			{
				int cur_track = fdc[ndrv_num].track;

				if (data < 0x80) fdc[ndrv_num].track++;
				else if (data > 0x80) fdc[ndrv_num].track--;

				if (fdc[ndrv_num].track < 0) fdc[ndrv_num].track = 0;
				if (fdc[ndrv_num].track > 255) fdc[ndrv_num].track = 255;

				if (cur_track != fdc[ndrv_num].track) {
					// fddseek sound turn on
					wav_play[0] = 1;
					wav_play_pos[0] = 0;
				}
			}
			break;
		case SIG_FLOPPY_TRACK_SIZE:
			if (config.fdd_type == 3 && fdcnum == 0) {
				// 26sectors max
				disk[ndrv_num]->track_size = TRACK_SIZE_8INCH_2D;
			} else {
				// 16sectors max
				disk[ndrv_num]->track_size = TRACK_SIZE_5INCH_2D;
			}
			fdc[ndrv_num].index = 0;
			break;
		case SIG_FLOPPY_CURRENTTRACK:
			fdc[ndrv_num].track = (data & mask);
			break;
		case SIG_FLOPPY_HEADLOAD:
			head_load = (data & mask);
			fdc[ndrv_num].head_loading = (head_load ? 6 : 0);
			OUT_DEBUG(_T("fdd %d sig HEADLOAD %x UL:%d"),drv_num,head_load,fdc[ndrv_num].head_loading);
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			cancel_my_events();
			reset();
			if (now_reset) {
				set_drive_select(IOPORT_USE_3FDD ? 0 : 1, 0);
				if (config.fdd_type == 3) set_drive_select(2, 0);
			}
			break;
	}

	if (IOPORT_USE_5FDD) {
		// for 5inch mini floppy
		fdcnum = (mask >> 16);
		mask &= 0xffff;
		uint8 nfdd5outreg = fdd5outreg[fdcnum];

		switch (id) {
			case SIG_FLOPPY_IRQ:
				irqflg = (data & mask) ? true : false;
				if (irqflg) {
					nfdd5outreg &= 0xfe;
				} else {
					nfdd5outreg |= 0x01;
				}
				break;
			case SIG_FLOPPY_DRQ:
				drqflg = (data & mask) ? true : false;
				if (drqflg) {
					nfdd5outreg &= 0x7f;
				} else {
					nfdd5outreg |= 0x80;
				}
				break;
		}

//		if (nfdd5outreg != fdd5outreg) {
//			OUT_DEBUG("fdd sig id:%04x d:%02x -> %02x"
//				,id,fdd5outreg[fdcnum],nfdd5outreg);
//		}

		// interrupt
		switch (id) {
		case SIG_FLOPPY_IRQ:
			if (irqflg != irqflgprev) {
				if (config.fdd_type == 3 || (drvsel[fdcnum] & 0x40) == 0) {	// nmi non mask
					register_my_event(EVENT_IRQ, 20);	// delay
				}
				irqflgprev = irqflg;
			}
			break;
		case SIG_FLOPPY_DRQ:
			if (drqflg != drqflgprev) {
				if (config.fdd_type == 3 && fdcnum == 0) { // halt release
//					register_my_event(EVENT_DRQ, 20);	// delay
					set_drq(drqflg);					// immediate
				}
				drqflgprev = drqflg;
			}
			break;
		}

		fdd5outreg[fdcnum] = nfdd5outreg;
	}
}

uint32 FLOPPY::read_signal(int id)
{
	uint32 data = 0;

	int fdcnum = (id >> 16);
	id &= 0xffff;

	int ndrv_num = drv_num[fdcnum];

	switch (id) {
		case SIG_FLOPPY_READ_ID:
			if(fdc[ndrv_num].index < 6) {
				data = disk[ndrv_num]->id[fdc[ndrv_num].index];
				fdc[ndrv_num].index++;
			}
			break;
		case SIG_FLOPPY_READ:
			if(fdc[ndrv_num].index < disk[ndrv_num]->sector_size) {
				data = disk[ndrv_num]->sector[fdc[ndrv_num].index];
				fdc[ndrv_num].index++;
			}
			break;
		case SIG_FLOPPY_READ_TRACK:
			if(fdc[ndrv_num].index < disk[ndrv_num]->track_size) {
				data = disk[ndrv_num]->track[fdc[ndrv_num].index];
				fdc[ndrv_num].index++;
			}
			break;
		case SIG_FLOPPY_WRITEPROTECT:
			data = (disk[ndrv_num]->write_protected || ignore_write) ? 1 : 0;
			break;
		case SIG_FLOPPY_READY:
			data = (disk[ndrv_num]->inserted && fdc[ndrv_num].ready >= 2) ? 1 : 0;
			break;
		case SIG_FLOPPY_TRACK0:
			data = (fdc[ndrv_num].track == 0) ? 1 : 0;
			break;
		case SIG_FLOPPY_INDEX:
			data = index_hole;
			break;
		case SIG_FLOPPY_DELETED:
			data = disk[ndrv_num]->deleted ? 1 : 0;
			break;
		case SIG_FLOPPY_CURRENTTRACK:
			data = fdc[ndrv_num].track;
			break;
		case SIG_FLOPPY_SECTOR_NUM:
			data = disk[ndrv_num]->sector_num;
			break;
		case SIG_FLOPPY_SECTOR_SIZE:
			data = disk[ndrv_num]->sector_size;
			break;
		case SIG_FLOPPY_TRACK_SIZE:
			data = disk[ndrv_num]->track_size;
			break;
	}

	return data;
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void FLOPPY::event_frame()
{
	for(int i=0; i<MAX_DRIVE; i++) {
		if (fdc[i].head_loading & 0xff) fdc[i].head_loading--;
	}
}

void FLOPPY::event_callback(int event_id, int err)
{
	int event_no = event_id;

	register_id[event_no] = -1;

	switch(event_no) {
		case EVENT_MOTOR_TIMEOUT:
			for(int i=0; i<MAX_DRIVE; i++) {
				// motor off
				motor(i, false);
			}
			break;
		case EVENT_INDEXHOLE_ON:
			// index hole on
			index_hole = 1;
			register_my_event(EVENT_INDEXHOLE_OFF, 300);	// index hole off
			break;
		case EVENT_INDEXHOLE_OFF:
			// index hole off
			index_hole = 0;
			for(int i=0; i<MAX_DRIVE; i++) {
				if (fdc[i].motor_warmup) {
					fdc[i].motor_warmup <<= 1;
					if (fdc[i].motor_warmup > 4) fdc[i].motor_warmup = 0;
					OUT_DEBUG(_T("fdd %d event MOTOR WARMUP %d"), i, fdc[i].motor_warmup);
				}
			}
			register_my_event(EVENT_INDEXHOLE_ON, delay_index_hole - 300);	// index hole on
			break;
		case EVENT_IRQ:
			cancel_my_event(EVENT_IRQ);
			set_irq(irqflg);
			break;
//		case EVENT_DRQ:
//			cancel_my_event(EVENT_DRQ);
//			set_drq(drqflg);
//			break;
		case EVENT_READY_ON_0:
		case EVENT_READY_ON_1:
		case EVENT_READY_ON_2:
		case EVENT_READY_ON_3:
		case EVENT_READY_ON_4:
		case EVENT_READY_ON_5:
		case EVENT_READY_ON_6:
		case EVENT_READY_ON_7:
			{
				int i = event_no - EVENT_READY_ON_0;
				fdc[i].ready <<= 1;
				OUT_DEBUG(_T("fdd %d event READY %d"), i, fdc[i].ready);
			}
			break;
		case EVENT_MOTOR_OFF_0:
		case EVENT_MOTOR_OFF_1:
		case EVENT_MOTOR_OFF_2:
		case EVENT_MOTOR_OFF_3:
		case EVENT_MOTOR_OFF_4:
		case EVENT_MOTOR_OFF_5:
		case EVENT_MOTOR_OFF_6:
		case EVENT_MOTOR_OFF_7:
			// motor off
			{
				int i = event_no - EVENT_MOTOR_OFF_0;
				motor(i, false);
			}
			break;
	}

}

// ----------------------------------------------------------------------------
/// @param [in] type 0:3inch 1:5inch 2:5inch ex
void FLOPPY::set_drive_select(int type, uint8 data)
{
	uint8 ndrv_num = 0;
	int fdcnum = (type == 2 ? 1 : 0);

	if (now_reset) {
		if (type != 0) {
			// 5inch
			data = 0;
		} else {
			// 3inch
			data = 0xff;
		}
	}

	// drive number
	if (type == 2) {
		// 5inch ex always use drive 2 or 3
		ndrv_num = (data & DRIVE_MASK) | 2;

	} else if (type != 0) {
		// 5inch
		ndrv_num = (data & DRIVE_MASK);

//		d_fdc5->write_signal(SIG_MB8866_DRIVEREG, ndrv_num, DRIVE_MASK);
	} else {
		// 3inch
		switch(data & 0x0f) {
			case 2:
				ndrv_num = 1;
				break;
			case 4:
				ndrv_num = 2;
				break;
			case 8:
				ndrv_num = 3;
				break;
			default:
				ndrv_num = 0;
				break;
		}
		ndrv_num &= DRIVE_MASK;
		if (drv_num[fdcnum] != ndrv_num) {
			// update status
			d_fdc3->write_signal(SIG_MC6843_UPDATESTATUS, 0, 0);
		}
	}
	drv_num[fdcnum] = ndrv_num;

	// set drive speed
	set_drive_speed();

	// side and motor
	if (type != 0) {
		// 5inch
		// side select
		set_disk_side(drv_num[fdcnum], (data & 0x10) ? 1 : 0);
//		d_fdc5->write_signal(SIG_MB8866_SIDEREG, data, 0x10);
		// density
		density = (data & 0x20) ? 1 : 0;

		// motor
		if ((data & 0x08) == 0 && (drvsel[fdcnum] & 0x08) != 0) {
			// motor on -> off
			OUT_DEBUG(_T("fdd %d MOTOR OFF REQUEST data:%02x drvsel:%02x"), drv_num, data, drvsel[fdcnum]);
			register_my_event(EVENT_MOTOR_OFF_0 + drv_num[fdcnum], DELAY_MOTOR_OFF_5);	// delay motor off
			data |= 0x08;
			drvsel[fdcnum] = (data & 0xff);
		} else if ((data & 0x08) != 0) {
			// motor on
			motor(drv_num[fdcnum], true);
			drvsel[fdcnum] = (data & 0xff);
		} else {
			// motor off
			OUT_DEBUG(_T("fdd %d MOTOR OFF FORCE data:%02x drvsel:%02x"), drv_num, data, drvsel[fdcnum]);
			motor(drv_num[fdcnum], false);
		}
//		d_fdc5->write_signal(SIG_MB8866_MOTOR, drvsel, 0x08);

	} else {
		// 3inch
		// motor
		if ((data & 0x80) == 0 && (drvsel[fdcnum] & 0x80) != 0) {
			// motor on -> off
			OUT_DEBUG(_T("fdd %d MOTOR OFF REQUEST data:%02x drvsel:%02x"), drv_num, data, drvsel[fdcnum]);
			register_my_event(EVENT_MOTOR_OFF_0 + drv_num[fdcnum], DELAY_MOTOR_OFF_3);	// delay motor off
			data |= 0x80;
			drvsel[fdcnum] = (data & 0xff);
		} else if ((data & 0x80) != 0) {
			// motor on
			motor(drv_num[fdcnum], true);
			drvsel[fdcnum] = (data & 0xff);
		} else {
			// motor off
			OUT_DEBUG(_T("fdd %d MOTOR OFF FORCE data:%02x drvsel:%02x"), drv_num, data, drvsel[fdcnum]);
			motor(drv_num[fdcnum], false);
		}

	}
}

void FLOPPY::motor(int drv, bool val)
{
	if (val) {
		// motor on
		register_my_event(EVENT_MOTOR_TIMEOUT, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)

		// motor sound on
		if (wav_play[1] == 0 && disk[drv]->inserted) {
			wav_play[1] = 1;
		}

		OUT_DEBUG(_T("fdd %d MOTOR ON REQUEST READY:%d WARMUP:%d"), drv, fdc[drv].ready, fdc[drv].motor_warmup);

		// ready on
		if (fdc[drv].ready == 0) {
			fdc[drv].ready = 1;
			register_my_event(EVENT_READY_ON_0 + drv, delay_ready_on);	// delay ready on (1s)
			if (disk[drv]->inserted) {
				fdc[drv].motor_warmup = 1;
			}
		}

		cancel_my_event(EVENT_MOTOR_OFF_0 + drv);

		OUT_DEBUG(_T("fdd %d MOTOR ON  READY:%d WARMUP:%d"), drv, fdc[drv].ready, fdc[drv].motor_warmup);
	} else {
		// motor off
		ignore_write = false;

		fdc[drv].ready = 0;
		fdc[drv].motor_warmup = 0;
		cancel_my_event(EVENT_MOTOR_OFF_0 + drv);

		// motor sound off when all drives stopped.
		wav_play[1] = 0;
		for(int i=0; i<MAX_DRIVE; i++) {
			wav_play[1] |= fdc[i].ready;
		}
		if (wav_play[1] < 2) {
			// reset motor flag when all drives stopped
			int fdcnum = (config.fdd_type == 3 && drv >= 2 ? 1 : 0);

			if (IOPORT_USE_3FDD) {
				// 3inch
				drvsel[fdcnum] &= 0x7f;
			} else if (IOPORT_USE_5FDD) {
				// 5inch
				drvsel[fdcnum] &= 0xf7;
			}
		}

		OUT_DEBUG(_T("fdd %d MOTOR OFF  SOUND:%d"), drv, wav_play[1]);
	}
}

void FLOPPY::set_drive_speed()
{
	if (config.fdd_type == 3) {
		delay_index_hole = DELAY_INDEX_HOLE_H;
	} else {
		delay_index_hole = DELAY_INDEX_HOLE;
	}
	if (config.fdd_type == 3) {
		delay_ready_on = DELAY_READY_ON_H;
	} else {
		delay_ready_on = DELAY_READY_ON;
	}
}

// ----------------------------------------------------------------------------
// media handler
// ----------------------------------------------------------------------------
bool FLOPPY::search_track(int channel)
{
	int fdcnum = (channel >> 16);
	return disk[drv_num[fdcnum]]->get_track(fdc[drv_num[fdcnum]].track, fdc[drv_num[fdcnum]].side);
}

bool FLOPPY::verify_track(int channel)
{
	// verify track number
	int fdcnum = (channel >> 16);
	for(int i = 0; i < disk[drv_num[fdcnum]]->sector_num; i++) {
		if(disk[drv_num[fdcnum]]->verify[i] != fdc[drv_num[fdcnum]].track) {
			return false;
		}
	}
	return true;
}

int FLOPPY::search_sector(int channel)
{
	int fdcnum = (channel >> 16);
	if (!disk[drv_num[fdcnum]]->get_sector(fdc[drv_num[fdcnum]].track, fdc[drv_num[fdcnum]].side, sectorcnt)) {
		return 1;	// SECTOR NOT FOUND
	}
	sectorcnt++;

	fdc[drv_num[fdcnum]].index = 0;

	if (disk[drv_num[fdcnum]]->status && !ignore_crc) {
		return 2;	// CRC ERROR
	}

	return 0;
}

int FLOPPY::search_sector(int channel, int sect)
{
	return search_sector(channel, sect, false, 0);
}

int FLOPPY::search_sector(int channel, int sect, bool compare_side, int side)
{
	int fdcnum = (channel >> 16);
	int status = 0;
	int sector_num = disk[drv_num[fdcnum]]->sector_num;
	sectorcnt = sect - 1;
	if(sectorcnt >= sector_num) {
		sectorcnt = 0;
	}
	int index = sectorcnt - 1;
	for(int i = 0; i < sector_num; i++) {
		// get sector
		index++;
		if(index >= sector_num) {
			index = 0;
		}
		if (!disk[drv_num[fdcnum]]->get_sector(fdc[drv_num[fdcnum]].track, fdc[drv_num[fdcnum]].side, index)) {
			continue;
		}

		// check sector in id field
		if(disk[drv_num[fdcnum]]->id[2] != sect) {
			continue;
		}
		// check side in id field
		if(compare_side && disk[drv_num[fdcnum]]->id[1] != side) {
			continue;
		}

		// check density
//		if(disk[drv_num]->density) {
//			continue;
//		}

		// sector found
		sectorcnt = index;

		fdc[drv_num[fdcnum]].index = 0;

		if (disk[drv_num[fdcnum]]->status && !ignore_crc) {
			status |= 2;	// CRC ERROR
		}

		return status;
	}
	return 1;	// SECTOR NOT FOUND
}

bool FLOPPY::make_track(int channel)
{
	int fdcnum = (channel >> 16);
	return disk[drv_num[fdcnum]]->make_track(fdc[drv_num[fdcnum]].track, fdc[drv_num[fdcnum]].side);
}

bool FLOPPY::parse_track(int channel)
{
	int fdcnum = (channel >> 16);
	return disk[drv_num[fdcnum]]->parse_track(fdc[drv_num[fdcnum]].track, fdc[drv_num[fdcnum]].side, density);
}

// ----------------------------------------------------------------------------
// irq / drq
// ----------------------------------------------------------------------------
void FLOPPY::set_irq(bool val)
{
	// NMI
	write_signals(&outputs_irq, val ? 0xffffffff : 0);

	if (config.fdd_type == 3 && val == true) {
		// clear HALT signal
		d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_FD_MASK);
	}
}

void FLOPPY::set_drq(bool val)
{
	// Always relase HALT signal
	write_signals(&outputs_drq, 0);
}

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------
bool FLOPPY::open_disk(int drv, const _TCHAR *path, int offset, uint32 flags)
{
	if(drv < MAX_DRIVE) {
		bool rc = disk[drv]->open(path, offset, flags);
		if (IOPORT_USE_5FDD) {
			// 5inch
			if (rc) {
				motor(drv, true);
				register_my_event(EVENT_MOTOR_OFF_0 + drv, DELAY_MOTOR_OFF_5S);	// delay motor off
			}
		}
		return rc;
	} else {
		return false;
	}
}

void FLOPPY::close_disk(int drv)
{
	if(drv < MAX_DRIVE) {
		disk[drv]->close();
		set_disk_side(drv, 0);
		motor(drv, false);
	}
}

int FLOPPY::change_disk(int drv)
{
	if(drv < MAX_DRIVE) {
		set_disk_side(drv, 1 - fdc[drv].side);
		motor(drv, false);
		return fdc[drv].side;
	} else {
		return 0;
	}
}

void FLOPPY::set_disk_side(int drv, int side)
{
	fdc[drv].side = side & 1;
	if (fdc[drv].side) {
		sidereg |= (1 << drv);
	} else {
		sidereg &= ~(1 << drv);
	}
}

int FLOPPY::get_disk_side(int drv)
{
	return fdc[drv].side;
}

bool FLOPPY::disk_inserted(int drv)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->inserted;
	}
	return false;
}

void FLOPPY::set_drive_type(int drv, uint8 type)
{
	if(drv < MAX_DRIVE) {
		disk[drv]->drive_type = type;
	}
}

uint8 FLOPPY::get_drive_type(int drv)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->drive_type;
	}
	return DRIVE_TYPE_UNK;
}

uint8 FLOPPY::fdc_status()
{
	return 0;
}

void FLOPPY::toggle_disk_write_protect(int drv)
{
	if(drv < MAX_DRIVE) {
		if (ignore_write) {
			ignore_write = false;
			disk[drv]->set_write_protect(true);
		}
		disk[drv]->set_write_protect(!(disk[drv]->write_protected));
	}
}

bool FLOPPY::disk_write_protected(int drv)
{
	if(drv < MAX_DRIVE) {
		return (disk[drv]->write_protected || ignore_write);
	}
	return true;
}

// ----------------------------------------------------------------------------
uint8 FLOPPY::get_drive_select()
{
	// for led indicator
	// b3-b0: drive select, b4-b7: 0:green led, 1:red led
	uint8 data = 0;

	if (IOPORT_USE_5FDD) {
		// 5inch
		for(int i=0; i<MAX_DRIVE; i++) {
			if (fdc[i].motor_warmup || (fdc[i].head_loading > 0 && fdc[i].ready >= 2)) {
				data |= ((1 << i) | (0x10 << i));
			}
		}
	} else if (IOPORT_USE_3FDD) {
		// 3inch
		if (drvsel[0] & 0x80) {
			data = (drvsel[0] & 0x0f) | ((sidereg & 0x0f) << 4);
		}
	}
	return data;
}

// ----------------------------------------------------------------------------
void FLOPPY::mix(int32* buffer, int cnt)
{
	int32 *buffer_start = buffer;
	int fdd_type = (IOPORT_USE_5FDD) ? 1 : 0;

	for(int typ=0; typ<2; typ++) {
		if (wav_play[typ] && wav_size[fdd_type][typ]) {
			buffer = buffer_start;
			for(int i=0; i<cnt; i++) {
				*buffer++ += ((int)wav_data[fdd_type][typ][wav_play_pos[typ]] - 128) * max_vol * wav_volume / 12800;
				*buffer++ += ((int)wav_data[fdd_type][typ][wav_play_pos[typ]] - 128) * max_vol * wav_volume / 12800;
				wav_play_pos[typ]++;

				if (wav_play_pos[typ] >= wav_size[fdd_type][typ]) {
					wav_play_pos[typ] = 0;
					if (typ == 0) {
						// seek sound play only one time.
						wav_play[typ] = 0;
						break;
					}
					// motor sound play loop.
//					else {
//					}
				}
			}
		}
	}
}

void FLOPPY::set_volume(int vol)
{
	wav_volume = vol;
}

void FLOPPY::init(int rate, int volume)
{
	sample_rate = rate;
	max_vol = volume;
	wav_volume = 0;

	// load wav file
	load_wav();
}

// ----------------------------------------------------------------------------
void FLOPPY::save_state(FILEIO *fp)
{
	struct vm_state_st_v2 vm_state;

	//
//	vm_state_ident.version = 1;
	vm_state_ident.version = 3;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	for(int i=0; i<22 && i<FLOPPY_MAX_EVENT; i++) {
		vm_state.register_id[i] = register_id[i];
	}
	vm_state.drv_num = drv_num[0];
	vm_state.drvsel  = drvsel[0];
	vm_state.sidereg = sidereg;
	vm_state.fdd5outreg = fdd5outreg[0];
	vm_state.irqflg  = irqflg ? 1 : 0;
	vm_state.drqflg  = drqflg ? 1 : 0;
	vm_state.drv_num2 = drv_num[1];
	vm_state.drvsel2  = drvsel[1];
	vm_state.fdd5outreg2 = fdd5outreg[1];

	fp->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fp->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void FLOPPY::load_state(FILEIO *fp)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st_v1 vm_state_v1;
	struct vm_state_st_v2 vm_state;

	FIND_STATE_CHUNK(fp, vm_state_i);

	// copy values
	if (vm_state_i.version == 1) {
		READ_STATE_CHUNK(fp, vm_state_i, vm_state_v1);

		for(int i=0; i<6; i++) {
			register_id[i] = vm_state_v1.register_id[i];
		}

		drv_num[0]  = vm_state_v1.drv_num;
		drvsel[0] = vm_state_v1.drvsel;
		sidereg = vm_state_v1.sidereg;
		fdd5outreg[0] = vm_state_v1.fdd5outreg;
		irqflg  = vm_state_v1.irqflg ? true : false;
		drqflg  = vm_state_v1.drqflg ? true : false;

		// cannot write to disk when load state from file.
		sectorcnt = 0;
		ignore_write = true;

	} else {
		READ_STATE_CHUNK(fp, vm_state_i, vm_state);

		for(int i=0; i<22 && i<FLOPPY_MAX_EVENT; i++) {
			register_id[i] = vm_state.register_id[i];
		}

		drv_num[0]  = vm_state.drv_num;
		drvsel[0] = vm_state.drvsel;
		sidereg = vm_state.sidereg;
		fdd5outreg[0] = vm_state.fdd5outreg;
		irqflg  = vm_state.irqflg ? true : false;
		drqflg  = vm_state.drqflg ? true : false;

		if (vm_state_ident.version >= 3) {
			drv_num[1] = vm_state.drv_num2;
			drvsel[1] = vm_state.drvsel2;
			fdd5outreg[1] = vm_state.fdd5outreg2;

		}

	}
	irqflgprev = irqflg;
	drqflgprev = drqflg;

	// sound off
	wav_play[0] = 0;
	wav_play[1] = 0;

	// fdd type is set on MEMORY class
	set_drive_speed();
}
