/** @file mb8866.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@par Origin
	mb8877.cpp
	@author Sasaji
	@date   2012.03.24 -

	@brief [ MB8866 modoki (same as MB8877/MB8876) ]
*/

#include "mb8866.h"
#include "floppy.h"
#include "../../config.h"

#ifdef _DEBUG_MB8866
#define OUT_DEBUG emu->out_debugf
#else
#define OUT_DEBUG dummyf
#endif

#define FDC_ST_BUSY		0x01	// busy
#define FDC_ST_INDEX		0x02	// index hole
#define FDC_ST_DRQ		0x02	// data request
#define FDC_ST_TRACK00		0x04	// track0
#define FDC_ST_LOSTDATA		0x04	// data lost
#define FDC_ST_CRCERR		0x08	// crc error
#define FDC_ST_SEEKERR		0x10	// seek error
#define FDC_ST_RECNFND		0x10	// sector not found
#define FDC_ST_HEADENG		0x20	// head engage
#define FDC_ST_RECTYPE		0x20	// record type
#define FDC_ST_WRITEFAULT	0x20	// write fault
#define FDC_ST_WRITEP		0x40	// write protectdc
#define FDC_ST_NOTREADY		0x80	// media not inserted

#define FDC_CMD_TYPE1		1
#define FDC_CMD_RD_SEC		2
#define FDC_CMD_RD_MSEC		3
#define FDC_CMD_WR_SEC		4
#define FDC_CMD_WR_MSEC		5
#define FDC_CMD_RD_ADDR		6
#define FDC_CMD_RD_TRK		7
#define FDC_CMD_WR_TRK		8
#define FDC_CMD_TYPE4		0x80

#define EVENT_SEEK		0
#define EVENT_SEEKEND		1
#define EVENT_SEARCH		2
#define EVENT_TYPE4		3
#define EVENT_MULTI1		4
#define EVENT_MULTI2		5
#define EVENT_LOST		6

#define DRIVE_MASK		(MAX_DRIVE - 1)

/// head loaded time 60ms
#define HEAD_LOADED_TIME	60000

/// seek stepping rate 6msec, 12msec, 20msec, 30msec
static const int seek_wait[4] = {6000, 12000, 20000, 30000};

#define CANCEL_EVENT(event) { \
	if(register_id[event] != -1) { \
		cancel_event(register_id[event]); \
		OUT_DEBUG(_T("FDC\tCancel EVENT:%d id:%d"), event, register_id[event]); \
		register_id[event] = -1; \
	} \
	if(event == EVENT_SEEK) { \
		now_seek = false; \
	} \
	if(event == EVENT_SEARCH) { \
		now_search = false; \
	} \
}
#define REGISTER_EVENT(event, wait) { \
	if(register_id[event] != -1) { \
		cancel_event(register_id[event]); \
		OUT_DEBUG(_T("FDC\tCancel EVENT:%d id:%d"), event, register_id[event]); \
		register_id[event] = -1; \
	} \
	register_event(this, (event << 8) | cmdtype, wait, false, &register_id[event]); \
	OUT_DEBUG(_T("FDC\tRegist EVENT:%d id:%d w:%d"), event, register_id[event], wait); \
	if(event == EVENT_SEEK) { \
		now_seek = true; \
	} \
	if(event == EVENT_SEARCH) { \
		now_search = true; \
	} \
}

void MB8866::initialize()
{
	// config
	ignore_crc = config.ignore_crc;

	// initialize fdc
	seektrk = 0;
	seekvct = true;
	status = cmdreg = trkreg = secreg = datareg = cmdtype = 0;
}

void MB8866::release()
{
}

void MB8866::reset()
{
	for(int i = 0; i < 7; i++) {
		register_id[i] = -1;
	}
	now_search = now_seek = after_seek = now_irq = false;

#ifdef _DEBUG_MB8866
	val0_prev = 0xffff;
	drq_prev = false;
#endif
}

void MB8866::cancel_my_events()
{
	for(int i = 0; i < 7; i++) {
		CANCEL_EVENT(i);
	}
}

void MB8866::update_config()
{
	ignore_crc = config.ignore_crc;
}

void MB8866::write_io8(uint32 addr, uint32 data)
{
	switch(addr & 3) {
	case 0:
		// command reg
#ifdef HAS_MB8876
		cmdreg = (~data) & 0xff;
#else
		cmdreg = data;
#endif
		process_cmd();
		break;
	case 1:
		// track reg
#ifdef HAS_MB8876
		trkreg = (~data) & 0xff;
#else
		trkreg = data;
#endif
		OUT_DEBUG(_T("FDCw\tTRACKREG=%d"), trkreg);
		if((status & FDC_ST_BUSY) && data_idx == 0) {
			// track reg is written after command starts
			if(cmdtype == FDC_CMD_RD_SEC || cmdtype == FDC_CMD_RD_MSEC || cmdtype == FDC_CMD_WR_SEC || cmdtype == FDC_CMD_WR_MSEC) {
				process_cmd();
			}
		}
		break;
	case 2:
		// sector reg
#ifdef HAS_MB8876
		secreg = (~data) & 0xff;
#else
		secreg = data;
#endif
		OUT_DEBUG(_T("FDCw\tSECREG=%d"), secreg);
		if((status & FDC_ST_BUSY) && data_idx == 0) {
			// sector reg is written after command starts
			if(cmdtype == FDC_CMD_RD_SEC || cmdtype == FDC_CMD_RD_MSEC || cmdtype == FDC_CMD_WR_SEC || cmdtype == FDC_CMD_WR_MSEC) {
				process_cmd();
			}
		}
		break;
	case 3:
		// data reg
#ifdef HAS_MB8876
		datareg = (~data) & 0xff;
#else
		datareg = data;
#endif
		if((status & FDC_ST_DRQ) && !now_search) {
			if(cmdtype == FDC_CMD_WR_SEC || cmdtype == FDC_CMD_WR_MSEC) {
				// write or multisector write
				if(d_fdd->read_signal(SIG_FLOPPY_WRITEPROTECT | channel)) {
					// write protect
//					status |= FDC_ST_WRITEFAULT;
					status |= FDC_ST_WRITEP;
					status &= ~FDC_ST_BUSY;
					status &= ~FDC_ST_DRQ;
//					cmdtype = 0;
					set_irq(true);
				} else {
					d_fdd->write_signal(SIG_FLOPPY_WRITE | channel, datareg, 0xff);
					// set deleted mark
					if (cmdreg & 1) {
						d_fdd->write_signal(SIG_FLOPPY_WRITEDELETE | channel, 1, 1);
					}
				}

				data_idx++;

				if(data_idx >= (int)d_fdd->read_signal(SIG_FLOPPY_SECTOR_SIZE | channel)) {
					if(cmdtype == FDC_CMD_WR_SEC) {
						// single sector
						OUT_DEBUG(_T("FDC\tEND OF SECTOR (%d bytes write)"), data_idx);
						status &= ~FDC_ST_BUSY;
//						cmdtype = 0;
						set_irq(true);
					}
					else {
						// multisector
						OUT_DEBUG(_T("FDC\tEND OF SECTOR (SEARCH NEXT)"));
						REGISTER_EVENT(EVENT_MULTI1, 30);
						REGISTER_EVENT(EVENT_MULTI2, 60);
					}
					status &= ~FDC_ST_DRQ;
				}
				d_fdd->write_signal(SIG_FLOPPY_ACCESS | channel, 1, 1);
			}
			else if(cmdtype == FDC_CMD_WR_TRK) {
				// write track
				if(d_fdd->read_signal(SIG_FLOPPY_WRITEPROTECT | channel)) {
					// write protect
//					status |= FDC_ST_WRITEFAULT;
					status |= FDC_ST_WRITEP;
					status &= ~FDC_ST_BUSY;
					status &= ~FDC_ST_DRQ;
//					cmdtype = 0;
					set_irq(true);
				} else {
					d_fdd->write_signal(SIG_FLOPPY_WRITE_TRACK | channel, datareg, 0xff);
				}

				data_idx++;

				if(data_idx >= (int)d_fdd->read_signal(SIG_FLOPPY_TRACK_SIZE | channel)) {
					status &= ~FDC_ST_BUSY;
					status &= ~FDC_ST_DRQ;
//					cmdtype = 0;
					set_irq(true);
				}
				d_fdd->write_signal(SIG_FLOPPY_ACCESS | channel, 1, 1);

				if(!(status & FDC_ST_DRQ)) {
					parse_track();
				}
			}
			if(!(status & FDC_ST_DRQ)) {
				set_drq(false);
			}
		}
		break;
	}
}

uint32 MB8866::read_io8(uint32 addr)
{
	uint32 val;

	switch(addr & 3) {
	case 0:
		// status reg
		if(cmdtype == FDC_CMD_TYPE4) {
			// now force interrupt
			if(d_fdd->read_signal(SIG_FLOPPY_READY | channel) == 0) {
				status = FDC_ST_NOTREADY;
			}
			else {
				// MZ-2500 RELICS invites STATUS = 0
				status = 0;
			}
			val = status;
		}
		else if(now_search) {
			// now sector search
			val = FDC_ST_BUSY;
		}
		else {
			// disk not inserted, motor stop
			if(d_fdd->read_signal(SIG_FLOPPY_READY | channel) == 0) {
				status |= FDC_ST_NOTREADY;
			}
			else {
				status &= ~FDC_ST_NOTREADY;
			}
			// write protected
			if(cmdtype == FDC_CMD_TYPE1 || cmdtype == FDC_CMD_WR_SEC || cmdtype == FDC_CMD_WR_MSEC || cmdtype == FDC_CMD_WR_TRK) {
				if(d_fdd->read_signal(SIG_FLOPPY_WRITEPROTECT | channel)) {
					status |= FDC_ST_WRITEP;
				}
				else {
					status &= ~FDC_ST_WRITEP;
				}
			}
			else {
				status &= ~FDC_ST_WRITEP;
			}
			// track0, index hole
			if(cmdtype == FDC_CMD_TYPE1) {
				if(d_fdd->read_signal(SIG_FLOPPY_TRACK0 | channel)) {
					status |= FDC_ST_TRACK00;
				}
				else {
					status &= ~FDC_ST_TRACK00;
				}
				if(!(status & FDC_ST_NOTREADY)) {
					if(d_fdd->read_signal(SIG_FLOPPY_INDEX | channel)) {
						status |= FDC_ST_INDEX;
					}
					else {
						status &= ~FDC_ST_INDEX;
					}
				}
			}
			// show busy a moment
			val = status;
			if(cmdtype == FDC_CMD_TYPE1 && !now_seek) {
				status &= ~FDC_ST_BUSY;
			}
		}

#ifdef _DEBUG_MB8866
//		// request cpu to output debug log
//		if(d_cpu) {
//			d_cpu->write_signal(SIG_CPU_DEBUG, 1, 1);
//		}
		val |= (now_irq ? 0x100 : 0);
		if (val != val0_prev) {
			OUT_DEBUG(_T("FDC\tSTATUS=%04x"), val);
			val0_prev = val;
		}
#endif

		// reset irq/drq
		if(!(status & FDC_ST_DRQ)) {
			set_drq(false);
		}
		set_irq(false);

#ifdef HAS_MB8876
		return (~val) & 0xff;
#else
		return val & 0xff;
#endif
	case 1:
		// track reg
		OUT_DEBUG(_T("FDC\tTRACKREG=%d"), trkreg);
#ifdef HAS_MB8876
		return (~trkreg) & 0xff;
#else
		return trkreg;
#endif
	case 2:
		// sector reg
		OUT_DEBUG(_T("FDC\tSECTORREG=%d"), secreg);
#ifdef HAS_MB8876
		return (~secreg) & 0xff;
#else
		return secreg;
#endif
	case 3:
		// data reg
		if((status & FDC_ST_DRQ) && !now_search) {
			if(cmdtype == FDC_CMD_RD_SEC || cmdtype == FDC_CMD_RD_MSEC) {
				// read or multisector read
				datareg = d_fdd->read_signal(SIG_FLOPPY_READ | channel);

				data_idx++;

				if(data_idx >= (int)d_fdd->read_signal(SIG_FLOPPY_SECTOR_SIZE | channel)) {
					if(cmdtype == FDC_CMD_RD_SEC) {
						// single sector
						OUT_DEBUG(_T("FDC\tEND OF SECTOR (%d bytes read)"), data_idx);
						status &= ~FDC_ST_BUSY;
//						cmdtype = 0;
						set_irq(true);
					}
					else {
						// multisector
						OUT_DEBUG(_T("FDC\tEND OF SECTOR (SEARCH NEXT)"));
						REGISTER_EVENT(EVENT_MULTI1, 30);
						REGISTER_EVENT(EVENT_MULTI2, 60);
					}
					status &= ~FDC_ST_DRQ;
				}
				d_fdd->write_signal(SIG_FLOPPY_ACCESS | channel, 1, 1);
			}
			else if(cmdtype == FDC_CMD_RD_ADDR) {
				// read address
				datareg = d_fdd->read_signal(SIG_FLOPPY_READ_ID | channel);

				data_idx++;

				if(data_idx >= 6) {
					status &= ~FDC_ST_BUSY;
					status &= ~FDC_ST_DRQ;
//					cmdtype = 0;
					set_irq(true);
				}
				d_fdd->write_signal(SIG_FLOPPY_ACCESS | channel, 1, 1);
			}
			else if(cmdtype == FDC_CMD_RD_TRK) {
				// read track
				datareg = d_fdd->read_signal(SIG_FLOPPY_READ_TRACK | channel);

				data_idx++;

				if(data_idx >= (int)d_fdd->read_signal(SIG_FLOPPY_TRACK_SIZE | channel)) {
					OUT_DEBUG(_T("FDC\tEND OF TRACK"));
					status &= ~FDC_ST_BUSY;
					status &= ~FDC_ST_DRQ;
					status |= FDC_ST_LOSTDATA;
//					cmdtype = 0;
					set_irq(true);
				}
				d_fdd->write_signal(SIG_FLOPPY_ACCESS | channel, 1, 1);
			}
			if(!(status & FDC_ST_DRQ)) {
				set_drq(false);
			}
		}
//		OUT_DEBUG(_T("FDC\tDATA=%2x"), datareg);
#ifdef HAS_MB8876
		return (~datareg) & 0xff;
#else
		return datareg;
#endif
	}
	return 0xff;
}

void MB8866::write_dma_io8(uint32 addr, uint32 data)
{
	write_io8(3, data);
}

uint32 MB8866::read_dma_io8(uint32 addr)
{
	return read_io8(3);
}

void MB8866::write_signal(int id, uint32 data, uint32 mask)
{
	if (id == SIG_CPU_RESET) {
		now_reset = (data & mask) ? true : false;
		cancel_my_events();
		reset();
	}
}

uint32 MB8866::read_signal(int ch)
{
	// get access status
	uint32 stat = 0;
	d_fdd->write_signal(SIG_FLOPPY_ACCESS | channel, 0, 1);

	return stat;
}

void MB8866::event_callback(int event_id, int err)
{
	int event = event_id >> 8;
	int cmd = event_id & 0xff;
	register_id[event] = -1;

	// cancel event if the command is finished or other command is executed
	if(cmd != cmdtype) {
		if(event == EVENT_SEEK) {
			now_seek = false;
		}
		else if(event == EVENT_SEARCH) {
			now_search = false;
		}
		return;
	}

	OUT_DEBUG(_T("FDC\tFire EVENT:%d id:%04x"),event,event_id);
	switch(event) {
	case EVENT_SEEK:
		{
			int seek = 0x80;
			if(seektrk > (int)d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel)) {
				seek = 0x7f; // plus
			}
			else if(seektrk < (int)d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel)) {
				seek = 0xff; // minus
			}
			d_fdd->write_signal(SIG_FLOPPY_STEP | channel, seek, 0xff);
			if(cmdreg & 0x10) {
				trkreg = d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel);
			}
			else if((cmdreg & 0xf0) == 0) {
				trkreg--;
			}
			if((uint32)seektrk == d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel)) {
				// auto update
				if((cmdreg & 0x10) || ((cmdreg & 0xf0) == 0)) {
					trkreg = d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel);
				}
				if((cmdreg & 0xf0) == 0) {
					datareg = 0;
				}
				status |= search_track();
				now_seek = false;
				set_irq(true);
			}
			else {
				REGISTER_EVENT(EVENT_SEEK, seek_wait[cmdreg & 3] + err);
			}
		}
		break;
	case EVENT_SEEKEND:
		if((uint32)seektrk == d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel)) {
			// auto update
			if((cmdreg & 0x10) || ((cmdreg & 0xf0) == 0)) {
				trkreg = d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel);
			}
			if((cmdreg & 0xf0) == 0) {
				datareg = 0;
			}
			status |= search_track();
			now_seek = false;
			CANCEL_EVENT(EVENT_SEEK);
			set_irq(true);
		}
		break;
	case EVENT_SEARCH:
		now_search = false;
		// start dma
		if(!(status & FDC_ST_RECNFND)) {
			status |= FDC_ST_DRQ;
			set_drq(true);
		}
		break;
	case EVENT_TYPE4:
		cmdtype = FDC_CMD_TYPE4;
		break;
	case EVENT_MULTI1:
		secreg++;
		break;
	case EVENT_MULTI2:
		if(cmdtype == FDC_CMD_RD_MSEC) {
			cmd_readdata();
		}
		else if(cmdtype == FDC_CMD_WR_MSEC) {
			cmd_writedata();
		}
		break;
	case EVENT_LOST:
		if(status & FDC_ST_BUSY) {
			status |= FDC_ST_LOSTDATA;
			status &= ~FDC_ST_BUSY;
			//status &= ~FDC_ST_DRQ;
			set_irq(true);
			//set_drq(false);
		}
		break;
	}
}

// ----------------------------------------------------------------------------
// command
// ----------------------------------------------------------------------------

void MB8866::process_cmd()
{
#ifdef _DEBUG_MB8866
	static const _TCHAR *cmdstr[0x10] = {
		_T("RESTORE "),	_T("SEEK    "),	_T("STEP    "),	_T("STEP    "),
		_T("STEP IN "),	_T("STEP IN "),	_T("STEP OUT"),	_T("STEP OUT"),
		_T("RD DATA "),	_T("RD DATA "),	_T("WR DATA "),	_T("WR DATA "),
		_T("RD ADDR "),	_T("FORCEINT"),	_T("RD TRACK"),	_T("WR TRACK")
	};
	OUT_DEBUG(_T("FDC\tCMD=%2xh (%s) DATA=%2xh TRK=%3d SEC=%2d"), cmdreg, cmdstr[cmdreg >> 4], datareg, trkreg, secreg);
#endif

	CANCEL_EVENT(EVENT_TYPE4);
	set_irq(false);

	switch(cmdreg & 0xf0) {
	// type-1
	case 0x00:
		cmd_restore();
		break;
	case 0x10:
		cmd_seek();
		break;
	case 0x20:
	case 0x30:
		cmd_step();
		break;
	case 0x40:
	case 0x50:
		cmd_stepin();
		break;
	case 0x60:
	case 0x70:
		cmd_stepout();
		break;
	// type-2
	case 0x80:
	case 0x90:
		cmd_readdata();
		break;
	case 0xa0:
	case 0xb0:
		cmd_writedata();
		break;
	// type-3
	case 0xc0:
		cmd_readaddr();
		break;
	case 0xe0:
		cmd_readtrack();
		break;
	case 0xf0:
		cmd_writetrack();
		break;
	// type-4
	case 0xd0:
		cmd_forceint();
		break;
	default:
		break;
	}
}

void MB8866::cmd_restore()
{
	// type-1 restore
	cmdtype = FDC_CMD_TYPE1;
	status = FDC_ST_BUSY;
	if (cmdreg & 8) {
		status |= FDC_ST_HEADENG;
	}
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, (cmdreg & 8) ? 1 : 0, 1);
	trkreg = 0xff;

	seektrk = 0;
	seekvct = true;

	REGISTER_EVENT(EVENT_SEEK, seek_wait[cmdreg & 3]);
	REGISTER_EVENT(EVENT_SEEKEND, 300);
}

void MB8866::cmd_seek()
{
	// type-1 seek
	cmdtype = FDC_CMD_TYPE1;
	status = FDC_ST_BUSY;
	if (cmdreg & 8) {
		status |= FDC_ST_HEADENG;
	}
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, (cmdreg & 8) ? 1 : 0, 1);
	after_seek = ((cmdreg & 8) ? false : true);

#if 0
	seektrk = fdc[drvreg].track + datareg - trkreg;
#else
	seektrk = datareg;
#endif
	seektrk = (seektrk > 83) ? 83 : (seektrk < 0) ? 0 : seektrk;
	seekvct = (datareg > trkreg) ? false : true;

	REGISTER_EVENT(EVENT_SEEK, seek_wait[cmdreg & 3]);
	REGISTER_EVENT(EVENT_SEEKEND, 300);
}

void MB8866::cmd_step()
{
	// type-1 step
	if(seekvct) {
		cmd_stepout();
	}
	else {
		cmd_stepin();
	}
}

void MB8866::cmd_stepin()
{
	// type-1 step in
	cmdtype = FDC_CMD_TYPE1;
	status = FDC_ST_BUSY;
	if (cmdreg & 8) {
		status |= FDC_ST_HEADENG;
	}
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, (cmdreg & 8) ? 1 : 0, 1);
	after_seek = ((cmdreg & 8) ? false : true);

	seektrk = d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel);
	seektrk = (seektrk < 83) ? seektrk + 1 : 83;
	seekvct = false;

	REGISTER_EVENT(EVENT_SEEK, seek_wait[cmdreg & 3]);
	REGISTER_EVENT(EVENT_SEEKEND, 300);
}

void MB8866::cmd_stepout()
{
	// type-1 step out
	cmdtype = FDC_CMD_TYPE1;
	status = FDC_ST_BUSY;
	if (cmdreg & 8) {
		status |= FDC_ST_HEADENG;
	}
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, (cmdreg & 8) ? 1 : 0, 1);
	after_seek = ((cmdreg & 8) ? false : true);

	seektrk = d_fdd->read_signal(SIG_FLOPPY_CURRENTTRACK | channel);
	seektrk = (seektrk > 0) ? seektrk - 1 : 0;
	seekvct = true;

	REGISTER_EVENT(EVENT_SEEK, seek_wait[cmdreg & 3]);
	REGISTER_EVENT(EVENT_SEEKEND, 300);
}

// wait 70msec to read/write data just after seek command is done
#define GET_SEARCH_TIME (after_seek ? (after_seek = false, HEAD_LOADED_TIME) : 200)

void MB8866::cmd_readdata()
{
	// type-2 read data
	d_fdd->write_signal(SIG_FLOPPY_CURRENTTRACK | channel, trkreg, 0xff);
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, 1, 1);

	cmdtype = (cmdreg & 0x10) ? FDC_CMD_RD_MSEC : FDC_CMD_RD_SEC;
	if(cmdreg & 2) {
		status = search_sector(((cmdreg & 8) ? 1 : 0), secreg, true);
	}
	else {
		status = search_sector(0, secreg, false);
	}
	if(!(status & FDC_ST_RECNFND)) {
//		status |= FDC_ST_DRQ | FDC_ST_BUSY;
		status |= FDC_ST_BUSY;
	}

	int time = GET_SEARCH_TIME;
	REGISTER_EVENT(EVENT_SEARCH, time);
	CANCEL_EVENT(EVENT_LOST);
	if(!(status & FDC_ST_RECNFND)) {
		REGISTER_EVENT(EVENT_LOST, time + 30000);
	}
}

void MB8866::cmd_writedata()
{
	// type-2 write data
	d_fdd->write_signal(SIG_FLOPPY_CURRENTTRACK | channel, trkreg, 0xff);
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, 1, 1);

	cmdtype = (cmdreg & 0x10) ? FDC_CMD_WR_MSEC : FDC_CMD_WR_SEC;
	if(cmdreg & 2) {
		status = search_sector(((cmdreg & 8) ? 1 : 0), secreg, true);
	}
	else {
		status = search_sector(0, secreg, false);
	}
	status &= ~FDC_ST_RECTYPE;
	if(!(status & FDC_ST_RECNFND)) {
//		status |= FDC_ST_DRQ | FDC_ST_BUSY;
		status |= FDC_ST_BUSY;
	}

	int time = GET_SEARCH_TIME;
	REGISTER_EVENT(EVENT_SEARCH, time);
	CANCEL_EVENT(EVENT_LOST);
	if(!(status & FDC_ST_RECNFND)) {
		REGISTER_EVENT(EVENT_LOST, time + 30000);
	}
}

void MB8866::cmd_readaddr()
{
	// type-3 read address
	d_fdd->write_signal(SIG_FLOPPY_CURRENTTRACK | channel, trkreg, 0xff);
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, 1, 1);

	cmdtype = FDC_CMD_RD_ADDR;
	status = search_addr();
	if(!(status & FDC_ST_RECNFND)) {
//		status |= FDC_ST_DRQ | FDC_ST_BUSY;
		status |= FDC_ST_BUSY;
	}

	int time = GET_SEARCH_TIME;
	REGISTER_EVENT(EVENT_SEARCH, time);
	CANCEL_EVENT(EVENT_LOST);
	if(!(status & FDC_ST_RECNFND)) {
		REGISTER_EVENT(EVENT_LOST, time + 10000);
	}
}

void MB8866::cmd_readtrack()
{
	// type-3 read track
	d_fdd->write_signal(SIG_FLOPPY_CURRENTTRACK | channel, trkreg, 0xff);
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, 1, 1);

	cmdtype = FDC_CMD_RD_TRK;
//	status = FDC_ST_DRQ | FDC_ST_BUSY;
	status = FDC_ST_BUSY;

	d_fdd->write_signal(SIG_FLOPPY_TRACK_SIZE | channel, 0x1800, 0xffff);
	make_track();

	int time = GET_SEARCH_TIME;
	REGISTER_EVENT(EVENT_SEARCH, time);
	REGISTER_EVENT(EVENT_LOST, time + 200000);
}

void MB8866::cmd_writetrack()
{
	// type-3 write track
	d_fdd->write_signal(SIG_FLOPPY_CURRENTTRACK | channel, trkreg, 0xff);
	d_fdd->write_signal(SIG_FLOPPY_HEADLOAD | channel, 1, 1);

	cmdtype = FDC_CMD_WR_TRK;
//	status = FDC_ST_DRQ | FDC_ST_BUSY;
	status = FDC_ST_BUSY;

	d_fdd->write_signal(SIG_FLOPPY_TRACK_SIZE | channel, 0x1800, 0xffff);
	data_idx = 0;

	int time = GET_SEARCH_TIME;
	REGISTER_EVENT(EVENT_SEARCH, time);
	REGISTER_EVENT(EVENT_LOST, time + 200000);
}

void MB8866::cmd_forceint()
{
	// type-4 force interrupt
#if 0
	if(!disk[drvreg]->inserted || !motor) {
		status = FDC_ST_NOTREADY | FDC_ST_HEADENG;
	}
	else {
		status = FDC_ST_HEADENG;
	}
	cmdtype = FDC_CMD_TYPE4;
#else
	if(cmdtype == 0 || cmdtype == 4) {
		status = 0;
		cmdtype = FDC_CMD_TYPE1;
	}
	status &= ~FDC_ST_BUSY;
#endif

	// force interrupt if bit0-bit3 is high
	if(cmdreg & 0x0f) {
		set_irq(true);
	}
	CANCEL_EVENT(EVENT_SEEK);
	CANCEL_EVENT(EVENT_SEEKEND);
	CANCEL_EVENT(EVENT_SEARCH);
	CANCEL_EVENT(EVENT_TYPE4);
	CANCEL_EVENT(EVENT_MULTI1);
	CANCEL_EVENT(EVENT_MULTI2);
	CANCEL_EVENT(EVENT_LOST);
	REGISTER_EVENT(EVENT_TYPE4, 100);
}

// ----------------------------------------------------------------------------
// media handler
// ----------------------------------------------------------------------------

uint8 MB8866::search_track()
{
	if(!d_fdd->search_track(channel)) {
		return FDC_ST_SEEKERR;
	}

	// verify track number
	if(!(cmdreg & 4)) {
		return 0;
	}
	if(!d_fdd->verify_track(channel)) {
		return FDC_ST_SEEKERR;
	}
	return 0;
}

uint8 MB8866::search_sector(int side, int sct, bool compare)
{
	// get track
	if(!d_fdd->search_track(channel)) {
		set_irq(true);
		return FDC_ST_RECNFND;
	}

	// scan sectors
	int stat = d_fdd->search_sector(channel, sct, compare, side);
	if (stat != 1) {
		data_idx = 0;
		return ((d_fdd->read_signal(SIG_FLOPPY_DELETED | channel)) ? FDC_ST_RECTYPE : 0) | ((status == 2) ? FDC_ST_CRCERR : 0);
	}
	// sector not found
	set_irq(true);
	return FDC_ST_RECNFND;
}

uint8 MB8866::search_addr()
{
	// get track
	if(!d_fdd->search_track(channel)) {
		set_irq(true);
		return FDC_ST_RECNFND;
	}

	// get sector
	int stat = d_fdd->search_sector(channel);
	if (stat != 1) {
		data_idx = 0;
		return ((status == 2) ? FDC_ST_CRCERR : 0);
	}

	// sector not found
	set_irq(true);
	return FDC_ST_RECNFND;
}

bool MB8866::make_track()
{
	return d_fdd->make_track(channel);
}

bool MB8866::parse_track()
{
	return d_fdd->parse_track(channel);
}

// ----------------------------------------------------------------------------
// irq / drq
// ----------------------------------------------------------------------------

void MB8866::set_irq(bool val)
{
	now_irq = val;
	write_signals(&outputs_irq, val ? 0xffffffff : 0);
#ifdef _DEBUG_MB8866
	if (val) {
		OUT_DEBUG(_T("FDC\tset_irq=true status:%02x"), status);
		val0_prev = 0xffff;
	}
#endif
}

void MB8866::set_drq(bool val)
{
	write_signals(&outputs_drq, val ? 0xffffffff : 0);
#ifdef _DEBUG_MB8866
	if (val != drq_prev) {
		OUT_DEBUG(_T("FDC\tset_drq=%s status:%02x"), val ? _T("true") : _T("false"), status);
		drq_prev = val;
	}
#endif
}

// ----------------------------------------------------------------------------
void MB8866::save_state(FILEIO *fp)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 1;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	for(int i=0; i<7; i++) {
		vm_state.register_id[i] = register_id[i];
	}
	vm_state.status = status;
	vm_state.cmdreg = cmdreg;
	vm_state.trkreg = trkreg;
	vm_state.secreg = secreg;
	vm_state.datareg = datareg;
	vm_state.cmdtype = cmdtype;
	vm_state.seektrk = seektrk;
	vm_state.data_idx = data_idx;
	vm_state.flags = (seekvct ? 8 : 0) | (now_search ? 4 : 0) | (after_seek ? 2 : 0) | (now_seek ? 1 : 0);

	fp->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fp->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void MB8866::load_state(FILEIO *fp)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fp, vm_state_i, vm_state);

	status = vm_state.status;
	cmdreg = vm_state.cmdreg;
	trkreg = vm_state.trkreg;
	secreg = vm_state.secreg;
	datareg = vm_state.datareg;
	cmdtype = vm_state.cmdtype;
	seektrk = vm_state.seektrk;
	data_idx = vm_state.data_idx;
	seekvct = (vm_state.flags & 8) ? true : false;
	now_search = (vm_state.flags & 4) ? true : false;
	after_seek = (vm_state.flags & 2) ? true : false;
	now_seek = (vm_state.flags & 1) ? true : false;

	// copy values
	for(int i=0; i<7; i++) {
		register_id[i] = vm_state.register_id[i];
	}

}
