/** @file mbs1.h

	HITACHI MB-S1 Emulator
	Skelton for retropc emulator

	@author Sasaji
	@date   2015.06.08 -

	@brief [ virtual machine ]
*/

#ifndef _MBS1_H_
#define _MBS1_H_

#define FRAME_SPLIT_NUM	1

#define DEVICE_NAME		"HITACHI MB-S1 model05"
#define CONFIG_NAME		"mbs1"
#define CLASS_NAME      "MBS1"
#define CONFIG_VERSION		0x01

// device informations for virtual machine
#define USE_EMU_INHERENT_SPEC

#define FRAMES_PER_10SECS	600
#define FRAMES_PER_SEC		60
#define LINES_PER_FRAME 	262
#define CHARS_PER_LINE		64

#define CLOCKS_1MHZ			1008000
#define CPU_CLOCKS			2016000
#define NUMBER_OF_CPUS		1
#define USE_CPU_REAL_MACHINE_CYCLE	1

#define USE_OPENGL
//#define SCREEN_WIDTH		640
#define SCREEN_WIDTH		768
//#define SCREEN_HEIGHT		480
#define SCREEN_HEIGHT		512
#define LIMIT_MIN_WINDOW_WIDTH		640
#define LIMIT_MIN_WINDOW_HEIGHT		400
#define MIN_WINDOW_WIDTH		640
#define MIN_WINDOW_HEIGHT		480
#define MAX_WINDOW_WIDTH		768
#define MAX_WINDOW_HEIGHT		512

// device informations for win32
#define USE_SPECIAL_RESET
#define USE_DATAREC
// #define USE_ALT_F10_KEY
#define USE_AUTO_KEY		3
#define USE_AUTO_KEY_CAPS
#define USE_SCANLINE
#define USE_DIPSWITCH
#define DIPSWITCH_DEFAULT 0x03
#define HAS_AY_3_8912
//#define USE_AUDIO_U8

#define USE_PRINTER
#define MAX_PRINTER		3
//#define USE_LIGHTPEN
//#define USE_MOUSE_ABSOLUTE
#define USE_JOYSTICK

#define USE_FD1
#define USE_FD2
#define USE_FD3
#define USE_FD4
#define HAS_MB8876
#define MAX_DRIVE		4
#define USE_DRIVE		4

#define USE_SOCKET
#define MAX_COMM		2

#define USE_KEY_RECORD

#define RESUME_FILE_HEADER "RESUME_MBS1"
#define RESUME_FILE_VERSION 1
#define RESUME_FILE_REVISION 1

#define KEYBIND_KEYS	130
#define KEYBIND_ASSIGN	2
#define KEYBIND_PRESETS	4

#define SIG_NMI_TRACE_MASK		0x01
#define SIG_NMI_TRAP_MASK		0x02
#define SIG_NMI_FD_MASK			0x04
#define SIG_NMI_KEYBREAK_MASK	0x08

#define SIG_IRQ_KEYBOARD_MASK	0x001
#define SIG_IRQ_MOUSE_MASK		0x002
#define SIG_IRQ_PIAA_MASK		0x004
#define SIG_IRQ_PIAB_MASK		0x008
#define SIG_IRQ_EXPIAA_MASK		0x010
#define SIG_IRQ_EXPIAB_MASK		0x020
#define SIG_IRQ_ACIA_MASK		0x040
#define SIG_IRQ_EXACIA_MASK		0x080
#define SIG_IRQ_9PSG_MASK		0x100

#define SIG_FIRQ_TIMER1_MASK	0x01
#define SIG_FIRQ_TIMER2_MASK	0x02
#define SIG_FIRQ_TIMER_MASK		0x03

#define SIG_HALT_PALETTE_MASK	0x01
#define SIG_HALT_FD_MASK		0x02

//#define _FDC_DEBUG_LOG

#include "../../common.h"
#include "registers.h"

class EMU;
class DEVICE;
class EVENT;

class HD46505;
class YM2203;
class MC6809;

class BOARD;
class PIA;
class ACIA;
class DISPLAY;
class KEYBOARD;
class MOUSE;
class MEMORY;
class SOUND;
class COMM;
class CMT;
class PRINTER;
class TIMER;
class KANJI;
class VIA;
class PSGC;
class PSG9C;
#ifdef USE_FD1
class MC6843;
class MB8866;
class FLOPPY;
#endif
class KEYRECORD;

/// for sleep/resume the machine
#pragma pack(1)
typedef struct vm_state_header_st {
	char header[16];
	uint16 version;
	uint16 revision;
	uint32 param;
	char reserved[8];
} vm_state_header_t;
#pragma pack()

/**
	@brief virtual machine for MBS1
*/
class VM
{
protected:
	EMU* emu;

	/// @name devices
	//@{
	EVENT* event;

	HD46505* crtc;
	YM2203* psg[2];
	YM2203* psg3[2];
	YM2203* psg9[3];
	MC6809* cpu;

	BOARD* board;
	PIA* pia;
	ACIA* acia;
	DISPLAY* display;
	KEYBOARD* key;
	MOUSE* mouse;
	MEMORY* memory;
	SOUND* sound;
	COMM* comm[MAX_COMM];
	CMT* cmt;
	PRINTER* printer[MAX_PRINTER];
	TIMER* timer;
	KANJI* kanji;
	PIA* psg3_pia;
	PSGC* psg3c;
	VIA* psg9_via;
	PSG9C* psg9c;
#ifdef USE_FD1
	MC6843* fdc3;
	MB8866* fdc5;
	MB8866* fdc5ex;
	FLOPPY* fdd;
#endif
	PIA* pia_ex;
	ACIA* acia_ex;

	DEVICE* dummy;
	//@}

	KEYRECORD* reckey;

public:
	DEVICE* first_device;
	DEVICE* last_device;

public:
	// ----------------------------------------
	// initialize
	// ----------------------------------------
	/// @name initialize
	//@{
	VM(EMU* parent_emu);
	~VM();
	//@}
	// ----------------------------------------
	// for emulation class
	// ----------------------------------------
	/// @name drive virtual machine
	//@{
	void run(int split_num);
	double frame_rate();
	bool now_skip();
	//@}
	/// @name draw screen
	//@{
	void set_display_size(int left, int top, int right, int bottom);
	void draw_screen();
	int  update_led();
	//@}
	/// @name sound generation
	//@{
	void initialize_sound(int rate, int samples);
#ifdef USE_AUDIO_U8
	uint8* create_sound(int* extra_frames, int samples);
#else
	int16* create_sound(int* extra_frames, int samples);
#endif
	//@}
	/// @name input event from emu
	//@{
	void key_down(int code);
	void key_up(int code);
	void system_key_down(int code);
	void system_key_up(int code);
	void set_mouse_position(int px, int py);
	//@}
	/// @name socket event from emu
	//@{
	void network_connected(int ch);
	void network_disconnected(int ch);
	void network_writeable(int ch);
	void network_readable(int ch);
	void network_accepted(int ch, int new_ch);
	uint8* get_sendbuffer(int ch, int* size);
	void inc_sendbuffer_ptr(int ch, int size);
	uint8* get_recvbuffer0(int ch, int* size0, int* size1);
	uint8* get_recvbuffer1(int ch);
	void inc_recvbuffer_ptr(int ch, int size);
	//@}
	/// @name control menu for user interface
	//@{
	void reset();
	void special_reset();
	void warm_reset(int onoff);
	void update_config();
	void change_dipswitch(int num);
	bool save_state(const _TCHAR* filename);
	bool load_state(const _TCHAR* filename);
#ifdef USE_KEY_RECORD
	bool play_reckey(const _TCHAR* filename);
	bool record_reckey(const _TCHAR* filename);
	void stop_reckey(bool stop_play = true, bool stop_record = true);
#endif
	//@}
	/// @name tape menu for user interface
	//@{
	bool play_datarec(const _TCHAR* file_path);
	bool rec_datarec(const _TCHAR* file_path);
	void close_datarec();
	void rewind_datarec();
	void fast_forward_datarec();
	void realmode_datarec();
	bool datarec_opened(bool play_mode);
	//@}
#ifdef USE_FD1
	/// @name floppy disk menu for user interface
	//@{
	bool open_disk(int drv, const _TCHAR* file_path, int offset, uint32 flags);
	void close_disk(int drv);
	int  change_disk(int drv);
	bool disk_inserted(int drv);
	int  get_disk_side(int drv);
	void toggle_disk_write_protect(int drv);
	bool disk_write_protected(int drv);
	//@}
#endif
	/// @name sound menu for user interface
	//@{
	void set_volume();
	//@}
	/// @name printer menu for user interface
	//@{
	bool save_printer(int dev, const _TCHAR* filename);
	void clear_printer(int dev);
	int  get_printer_buffer_size(int dev);
	uint8* get_printer_buffer(int dev);
	void enable_printer_direct(int dev);
	bool print_printer(int dev);
	//@}
	/// @name comm menu for user interface
	//@{
	void enable_comm_server(int dev);
	void enable_comm_connect(int dev);
	//@}
	/// @name options menu for user interface
	//@{
	void save_keybind();
	//@}

	// ----------------------------------------
	// for each device
	// ----------------------------------------
	/// @name event callbacks
	//@{
	void register_event(DEVICE* device, int event_id, int usec, bool loop, int* register_id);
	void register_event_by_clock(DEVICE* device, int event_id, int clock, bool loop, int* register_id);
	void cancel_event(int register_id);
	void register_frame_event(DEVICE* dev);
	void register_vline_event(DEVICE* dev);
	void set_lines_per_frame(int lines);
	//@}
	/// @name clock
	//@{
	uint64 current_clock();
	uint64 passed_clock(uint64 prev);
	uint32 get_pc();
	//@}
	/// @name get devices
	//@{
	DEVICE* get_device(int id);
	DEVICE* get_device(char *name, char *identifier);
	//@}
};

#endif
